@@ -721,6 +721,7 @@ enum act_return http_action_store_cache(struct act_rule *rule, struct proxy *px,
 	if (IS_HTX_STRM(s)) {
 		struct htx *htx = htxbuf(&s->res.buf);
 		struct http_hdr_ctx ctx;
+		size_t hdrs_len = 0;
 		int32_t pos;
 
 		/* Do not cache too big objects. */
@@ -755,11 +756,17 @@ enum act_return http_action_store_cache(struct act_rule *rule, struct proxy *px,
 			enum htx_blk_type type = htx_get_blk_type(blk);
 			uint32_t sz = htx_get_blksz(blk);
 
+			hdrs_len += sizeof(*blk) + sz;
 			chunk_memcat(&trash, (char *)&blk->info, sizeof(blk->info));
 			chunk_memcat(&trash, htx_get_blk_ptr(htx, blk), sz);
 			if (type == HTX_BLK_EOH)
 				break;
 		}
+
+		/* Do not cache objects if the headers are too big. */
+		if (hdrs_len > htx->size - global.tune.maxrewrite)
+			goto out;
+
 	}
 	else {
 		struct hdr_ctx ctx;
@@ -789,6 +796,10 @@ enum act_return http_action_store_cache(struct act_rule *rule, struct proxy *px,
 			}
 			http_remove_header2(msg, &txn->hdr_idx, &ctx);
 		}
+
+		/* Do not cache objects if the headers are too big. */
+		if (msg->sov > c_size(txn->rsp.chn) - global.tune.maxrewrite)
+			goto out;
 	}
 
 	shctx_lock(shctx);
@@ -899,6 +910,7 @@ static unsigned int htx_cache_dump_blk(struct appctx *appctx, struct htx *htx, e
 	struct cache_flt_conf *cconf = appctx->rule->arg.act.p[0];
 	struct shared_context *shctx = shctx_ptr(cconf->c.cache);
 	struct htx_blk *blk;
+	char *ptr;
 	unsigned int max, total;
 	uint32_t blksz;
 
@@ -917,12 +929,14 @@ static unsigned int htx_cache_dump_blk(struct appctx *appctx, struct htx *htx, e
 
 	blk->info = info;
 	total = 4;
+	ptr = htx_get_blk_ptr(htx, blk);
 	while (blksz) {
 		max = MIN(blksz, shctx->block_size - offset);
-		memcpy(htx_get_blk_ptr(htx, blk), (const char *)shblk->data + offset, max);
+		memcpy(ptr, (const char *)shblk->data + offset, max);
 		offset += max;
 		blksz  -= max;
 		total  += max;
+		ptr    += max;
 		if (blksz || offset == shctx->block_size) {
 			shblk = LIST_NEXT(&shblk->list, typeof(shblk), list);
 			offset = 0;
