@@ -25,7 +25,7 @@
  * OTHER DEALINGS IN THE SOFTWARE.
  */
 
-#include <stdint.h>
+#include <inttypes.h>
 #include <common/config.h>
 #include <common/h2.h>
 #include <common/http-hdr.h>
@@ -555,6 +555,8 @@ static struct htx_sl *h2_prepare_htx_reqline(uint32_t fields, struct ist *phdr,
 		goto fail;
 
 	sl->info.req.meth = find_http_meth(phdr[H2_PHDR_IDX_METH].ptr, phdr[H2_PHDR_IDX_METH].len);
+	sl->flags |= HTX_SL_F_HAS_SCHM;
+	sl->flags |= (isteqi(phdr[H2_PHDR_IDX_SCHM], ist("http")) ? HTX_SL_F_SCHM_HTTP : HTX_SL_F_SCHM_HTTPS);
 	return sl;
  fail:
 	return NULL;
@@ -590,6 +592,7 @@ int h2_make_htx_request(struct http_hdr *list, struct htx *htx, unsigned int *ms
 	int phdr;
 	int ret;
 	int i;
+	uint32_t used = htx_used_space(htx);
 	struct htx_sl *sl = NULL;
 	unsigned int sl_flags = 0;
 
@@ -733,7 +736,7 @@ int h2_make_htx_request(struct http_hdr *list, struct htx *htx, unsigned int *ms
 			if (tl > fs)
 				goto fail;
 
-			htx_set_blk_value_len(blk, tl);
+			htx_change_blk_value_len(htx, blk, tl);
 			*(char *)(htx_get_blk_ptr(htx, blk) + bs + 0) = ';';
 			*(char *)(htx_get_blk_ptr(htx, blk) + bs + 1) = ' ';
 			memcpy(htx_get_blk_ptr(htx, blk) + bs + 2, list[ck].v.ptr, vl);
@@ -745,6 +748,9 @@ int h2_make_htx_request(struct http_hdr *list, struct htx *htx, unsigned int *ms
 	/* now send the end of headers marker */
 	htx_add_endof(htx, HTX_BLK_EOH);
 
+	/* Set bytes used in the HTX mesage for the headers now */
+	sl->hdrs_bytes = htx_used_space(htx) - used;
+
 	ret = 1;
 	return ret;
 
@@ -830,6 +836,7 @@ int h2_make_htx_response(struct http_hdr *list, struct htx *htx, unsigned int *m
 	int phdr;
 	int ret;
 	int i;
+	uint32_t used = htx_used_space(htx);
 	struct htx_sl *sl = NULL;
 	unsigned int sl_flags = 0;
 
@@ -928,6 +935,9 @@ int h2_make_htx_response(struct http_hdr *list, struct htx *htx, unsigned int *m
 	/* now send the end of headers marker */
 	htx_add_endof(htx, HTX_BLK_EOH);
 
+	/* Set bytes used in the HTX mesage for the headers now */
+	sl->hdrs_bytes = htx_used_space(htx) - used;
+
 	ret = 1;
 	return ret;
 
@@ -935,13 +945,11 @@ int h2_make_htx_response(struct http_hdr *list, struct htx *htx, unsigned int *m
 	return -1;
 }
 
-/* Takes an H2 headers list <list> terminated by a name being <NULL,0> and
- * emits the equivalent HTX trailers block not including the empty line. The
- * output contents are emitted in <htx>, and a positive value is returned if
- * some bytes were emitted. In case of error, a negative error code is
- * returned. The caller must have verified that the message in the buffer is
- * compatible with receipt of trailers. Note that for now the HTX trailers
- * block is in fact an H1 block and it must contain the trailing CRLF.
+/* Takes an H2 headers list <list> terminated by a name being <NULL,0> and emits
+ * the equivalent HTX trailers blocks. The output contents are emitted in <htx>,
+ * and a positive value is returned if some bytes were emitted. In case of
+ * error, a negative error code is returned. The caller must have verified that
+ * the message in the buffer is compatible with receipt of trailers.
  *
  * The headers list <list> must be composed of :
  *   - n.name != NULL, n.len  > 0 : literal header name
@@ -953,13 +961,9 @@ int h2_make_htx_response(struct http_hdr *list, struct htx *htx, unsigned int *m
  */
 int h2_make_htx_trailers(struct http_hdr *list, struct htx *htx)
 {
-	struct htx_blk *blk;
-	char *out;
 	uint32_t idx;
-	int len;
 	int i;
 
-	len = 2; // CRLF
 	for (idx = 0; list[idx].n.len != 0; idx++) {
 		if (!list[idx].n.ptr) {
 			/* This is an indexed pseudo-header (RFC7540#8.1.2.1) */
@@ -987,29 +991,13 @@ int h2_make_htx_trailers(struct http_hdr *list, struct htx *htx)
 		    isteq(list[idx].n, ist("transfer-encoding")))
 			goto fail;
 
-		len += list[idx].n.len + 2 + list[idx].v.len + 2;
+		if (!htx_add_trailer(htx, list[idx].n, list[idx].v))
+			goto fail;
 	}
 
-	blk = htx_add_blk_type_size(htx, HTX_BLK_TLR, len);
-	if (!blk)
+	if (!htx_add_endof(htx, HTX_BLK_EOT))
 		goto fail;
 
-	out = htx_get_blk_ptr(htx, blk);
-	for (idx = 0; list[idx].n.len != 0; idx++) {
-		/* copy "name: value" */
-		memcpy(out, list[idx].n.ptr, list[idx].n.len);
-		out += list[idx].n.len;
-		*(out++) = ':';
-		*(out++) = ' ';
-
-		memcpy(out, list[idx].v.ptr, list[idx].v.len);
-		out += list[idx].v.len;
-		*(out++) = '\r';
-		*(out++) = '\n';
-	}
-	*(out++) = '\r';
-	*(out++) = '\n';
-
 	return 1;
 
  fail:
